/***************************************************************************
 *   Copyright (C) 2006 by BardaK   *
 *   bardak@BardaX   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <netinet/in.h>
#include "peerinfo.h"
#include "trackerUtils.h"

void addTracker(struct PeerInfo* pi, struct Tracker* t){
	if(pi->trackersLength == 0){
		pi->trackers = (struct Tracker**)malloc(sizeof(struct Tracker*));
		if(!pi->trackers){			
			fprintf(stderr, "ERROR: No memory for PeerInfo's trackers\n");
			exit(1);
		}
		pi->trackersLength = 1;
		pi->trackers[0] = t;
	}else{
		int i = 0;
		for(; i < pi->trackersLength; i++)
			if(pi->trackers[i] == t)
				break;
		if(i == pi->trackersLength){
			pi->trackers = (struct Tracker**)realloc(pi->trackers, sizeof(struct Tracker*)*(pi->trackersLength +1));
			if(!pi->trackers){	
				fprintf(stderr, "ERROR: No memory for PeerInfo's trackers\n");
				exit(1);
			}
			pi->trackers[pi->trackersLength] = t;
			pi->trackersLength ++;
		}
	}
}

void addPeerInfo(struct PeerInfo** dest, struct PeerInfo* source){
	if(*dest == 0){
		*dest = source;
		return;
	}
	struct PeerInfo* destination = *dest;
	int cmpResult = comparePeerInfo(destination, source);
	if(cmpResult > 0){
		if(destination->left == 0)
			destination->left = source;
		else
			addPeerInfo(&(destination->left), source);
	}else if(cmpResult <0){
		if(destination->right == 0)
			destination->right = source;
		else
			addPeerInfo(&(destination->right), source);
	}else if(!cmpResult){
		int i = 0;
		for(; i < source->trackersLength; i++)
			addTracker(destination, source->trackers[i]);
		if(!destination->peer->peerid && source->peer->peerid)
			strcpy(destination->peer->peerid, source->peer->peerid);;
		free(source);
	}
	
	return;
}

void addPeerToPeerInfo(struct PeerInfo** destination, struct Peer* peer, struct Tracker* tracker){
	struct PeerInfo * newpi = (struct PeerInfo*)malloc(sizeof(struct PeerInfo));
	memset(newpi, 0, sizeof(struct PeerInfo));
	newpi->peer = peer;
	addTracker(newpi, tracker);
	addPeerInfo(destination, newpi);	
}

int comparePeerInfo(struct PeerInfo*first, struct PeerInfo*second){
	if(first->peer->ip > second->peer->ip)
		return 1;
	if(first->peer->ip < second->peer->ip)
		return -1;
	if(first->peer->port > second->peer->port)
		return 1;
	if(first->peer->port < second->peer->port)
		return -1;
	return 0;
}

void printCSVStructure(FILE* fd, struct PeerInfo* pi, char compact, char trackers){
	if(!pi)return;
	struct in_addr ip;
	ip.s_addr = pi->peer->ip;			
	fprintf(fd, "\"%s\",%d", inet_ntoa(ip), ntohs(pi->peer->port));
	if(!compact){
		char peerid[61];
		memset(peerid,0,61);
		urlEncode(peerid, pi->peer->peerid,20);
		fprintf(fd, ",\"%s\"",peerid);
	}
	if(trackers){
		int i =0;
		for(; i < pi->trackersLength; i ++){
			fprintf(fd, ",\"");
			printTrackerAddress(fd,pi->trackers[i]);
			fprintf(fd, "\"");
		}
	}
	fprintf(fd, "\n");	
	if(pi->left)
		printCSVStructure(fd, pi->left, compact, trackers);
	if(pi->right)
		printCSVStructure(fd, pi->right, compact, trackers);
}

