/***************************************************************************
 *   Copyright (C) 2006 by BardaK   *
 *   bardak@BardaX   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <openssl/sha.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <mcheck.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <signal.h>
#include <fcntl.h>
#include <netinet/tcp.h>
#include "trackerUtils.h"
#include "bencodeUtils.h"
#include "peerinfo.h"
#define PEER_ID "-AZ2060-White Rabbit"
#define CLIENTPORT  6929
#define BARLENGTH 40
#define REQNUM 100
#define NUMWANT 1000
#define MAXSYNCNT 1
#define RESPONSE_SIZE 100000

#define LOG_MODE_TRACKER 1
#define LOG_MODE_REQUEST 2
#define LOG_MODE_IPADDRS  4

#define OPT_ANNOUNCE_START 1
#define OPT_SCRAPE 2
#define OPT_ANNOUNCE_STOP 4
#define OPT_COMPACT 8



#ifndef TCP_SYNCNT
#define TCP_SYNCNT 7
#endif
 
char * metaInfoFile, *hashEncoded, logmode;
unsigned long metaInfoFileLength;
char description = 1, compact = 1;
FILE*logFD=0;


struct Request{
	struct Tracker * tracker;
	unsigned int num;
	unsigned int newpeers;
	unsigned int total;
	double time;
};

struct Tracker * trackers;
unsigned int trackersNum = 0;
struct Request * requests;
unsigned int requestsNum = 0;



inline char nibbleToHexChar(char value,int pos) {
    value >>= 4*pos;
    value &= 0xF;
    if(value < 10)
        return '0' + value;
    else
        return 'A' + value - 10;
}

char* urlEncode(char* encodedMessage, const char* content, int length) {
    int i = 0;
    memset(encodedMessage, 0, 3*length +1);
    char* destPtr = encodedMessage;
    for (i = 0; i < length; i++) {
	    if(isdigit(content[i]) 
		      || islower(content[i]) 
		      || content[i] == '-'
		      || content[i] == '$'
		      || content[i] == '_'
		      || content[i] == '.'
		      || content[i] == '+'
		      || content[i] == '!'
		      || content[i] == '*'
		      || content[i] == '\''
		      || content[i] == '('
		      || content[i] == ')') {
            *(destPtr++)=content[i];
        } else {
            *(destPtr++)='%';
            *(destPtr++)=nibbleToHexChar(content[i],1);
            *(destPtr++)=nibbleToHexChar(content[i],0);
        }
    }
    return encodedMessage;
}


char* getInfoHash(char* destination) {
    char* pointer = metaInfoFile;
    while(pointer < metaInfoFile+metaInfoFileLength && strncmp(pointer,"4:info",6))
        pointer++;
    if(pointer >= metaInfoFile+metaInfoFileLength) return 0;
    pointer+=6;
    unsigned long infochars = bencodeSkipChars(pointer);
    char * digest = destination;
    SHA1(pointer,infochars,digest);
    //SHA1Context ctx;
    //SHA1Input(&ctx,buffer, infochars);
    //SHA1Result(&ctx, digest);
    return digest;
}


int makeGETRequest(char* get, struct Tracker* tracker, int option){
	memset(get, 0,500);
    /*
     * Init path for http request
     */
    char* path = tracker->announce, *scrape;
    if(option & OPT_SCRAPE) {
        scrape =(char*)malloc(strlen(tracker->announce) -1);
        getScrapePath(tracker, scrape);
        path = scrape;
    }
    sprintf(get, "GET %s?info_hash=%s", path, hashEncoded);
    if(option & OPT_SCRAPE){
	    free(path);
	    strcat(get, " HTTP/1.0\r\n");	    
	    sprintf(get+strlen(get), "Host: %s:%d\r\n\r\n", tracker->hostname, tracker->port);
	    return 1;
    }
    /*
     * Init peerId value
     */
    char * peer_id = (char*)malloc(3*20+1);
    if(!peer_id) {
        fprintf(stderr, "ERROR: No memory for peer_id string\n");
        return 0;
    }
    peer_id[3*20]=0;
    memset(peer_id,3*20,' ');
    urlEncode(peer_id,PEER_ID, strlen(PEER_ID));
    /*
    * Init trackerid value
    */
    char* trackerid=0;
    if(tracker->trackerid){
	    	trackerid = (char*)malloc(255);
	        sprintf(trackerid, "&trackerid=%s", tracker->trackerid);
    } 
    /*
     * Concatenate rest of parameters in get pointer
    */
    int stop = (option & OPT_ANNOUNCE_STOP) == OPT_ANNOUNCE_STOP;
    sprintf(get+strlen(get), "&peer_id=%s&port=%d&uploaded=0&downloaded=0&left=1&event=%s&numwant=%d%s%s HTTP/1.1\r\n", peer_id,CLIENTPORT, stop?"stopped":"started",stop?0:NUMWANT, (option & OPT_COMPACT)?"&compact=1":"", trackerid?trackerid:"");
    sprintf(get+strlen(get), "Host: %s:%d\r\n\r\n", tracker->hostname, tracker->port);
    free(peer_id);
    if(!trackerid)
	    free(trackerid);
    return 1;
}


int initTrackerAddress(struct sockaddr_in *remote, struct Tracker* tracker) {
    remote->sin_family = AF_INET;
    remote->sin_port=htons(tracker->port);
    /*
    * Obtain Ip address from hostname
    */
    struct hostent * trackerInfo = gethostbyname(tracker->hostname);
    if(!trackerInfo) {
        fprintf(stderr,"WARNING: Cannot obtain information about host: %s\n", tracker->hostname);
        return 0;
    }
    if(trackerInfo->h_addrtype != AF_INET) {
        fprintf(stderr, "WARNING: Not a IP host: %s\n",tracker->hostname);
        return 0;
    }
    if(*(trackerInfo->h_addr_list))
        remote->sin_addr.s_addr = ((struct in_addr *)*(trackerInfo->h_addr_list))->s_addr;
    else {
        fprintf(stderr, "WARNING: No IP address found from alias\n");
        return 0;
    }
    return 1;
}


void getTrackerURL() {
    const char* pointer = metaInfoFile;
    const char* const end = metaInfoFile+metaInfoFileLength;
    char* tempBuff = (char*)malloc(20);
    const char* announce;
    unsigned int numchars = 0;
    while(pointer < end){
	    //Searching for name key
	    if(startsWith(pointer,"4:name",end-pointer)){
	    //if(pointer < end -6 && !strncmp(pointer, "4:name",6)){
		    pointer+=6; 
		    numchars = strchr(pointer, ':')-pointer;
		    tempBuff[numchars]=0;
		    strncpy(tempBuff, pointer, numchars);
		    pointer += numchars + 1;
		    numchars = atoi(tempBuff);
		    printf("Torrent name: ");
		    fwrite(pointer, numchars, 1, stdout);
		    printf("\n");
		    pointer+=numchars;
    	    //Searching for announce key
	    }else if(startsWith(pointer, "8:announce", end-pointer)){
	    //}else if(pointer < end -10 && !strncmp(pointer, "8:announce",10)){
		    pointer+=10;
		    numchars = strchr(pointer, ':')-pointer;
		    tempBuff[numchars]=0;
		    strncpy(tempBuff, pointer, numchars);
		    pointer += numchars + 1;
		    numchars = atoi(tempBuff);
		    trackers = newTracker(trackers,&trackersNum);
		    initTracker(trackers,pointer, numchars);
		    announce=pointer;
		    pointer+=numchars;
	    }else if(startsWith(pointer, "13:announce-list",end-pointer)){
		    pointer+=16;
		    pointer++;
		    trackersNum=0;
		    freeTracker(trackers);
		    unsigned int offset = 0;
		    while(*pointer == 'l') {
			    pointer++;
jump:
			    numchars = strchr(pointer, ':')-pointer;
            		    tempBuff[numchars]=0;
			    strncpy(tempBuff, pointer, numchars);
			    pointer += numchars +1;
			    //if(offset != 1 || strncmp(pointer, announce, atoi(tempBuff))) {
				    trackers= newTracker(trackers,&trackersNum);
				    initTracker((trackers+offset),pointer,atoi(tempBuff));
				    offset++;
			    //}
			    pointer += atoi(tempBuff);
			    if(*pointer == 'e')
				    pointer++;
			    else if(isdigit(*pointer))
				    goto jump;
			    else
				    printf("BUG: Something wrong in the tracker url process \n");
		    }
	    }else if(startsWith(pointer, "4:info",end-pointer)){
		    pointer += 6;
		    pointer += bencodeSkipChars(pointer);
	    }else 
		    pointer++;
    }
    free(tempBuff);
}





int parseResponse( const char*  httpResponse, unsigned int size, struct Tracker* tracker, char option) {
    const char* pointer = httpResponse;
    const char* const end = httpResponse+size;
    int seed=-1;
    int leech=-1;
    int ret = 0;
    while(pointer < end){	 
	    if(startsWith(pointer, "HTTP/1.",end-pointer)){
		    pointer+=8;
		    while(*pointer==' ')
			    pointer++;
		    char num[4];
		    strncpy(num, pointer, 3);
		    num[3]=0;
		    tracker->httpCode = atoi(num);
		    while(*pointer != '\r' && *pointer != '\n')
			    pointer++;
		    if(*pointer == '\n')pointer++;
		    if(tracker->httpCode == 200){
			    if(option & OPT_SCRAPE)
				    setScrape(tracker,1);
			    else
				    setAnnounce(tracker,1);
		    }else{
			    if(option & OPT_SCRAPE)
				    setScrape(tracker,0);
			    else
				    setAnnounce(tracker,0);
			    break;
		    }	
	    }else if(startsWith(pointer, "14:failure reason", end-pointer)){
		    pointer += 17;
		    int numchars = strchr(pointer,':')-pointer;
		    if (numchars < 0) return 0;
		    char*  tempBuff = (char*)malloc(numchars+1);
		    tempBuff[numchars]=0;
		    strncpy(tempBuff, pointer, numchars);
		    pointer += numchars +1;
		    char * failure = (char*) calloc(1,atoi(tempBuff));
		    strncpy(failure,pointer, atoi(tempBuff));
		    fprintf(stderr,"FAILURE: %s\n", failure);
		    if(option & OPT_COMPACT) 
			    tracker->failure = failure;
		    if(option & OPT_SCRAPE)
			    setScrape(tracker,0);
		    else
			    setAnnounce(tracker,0);
		    return -1;
	    }else if(!(option & OPT_SCRAPE) & startsWith(pointer, "5:peers", end-pointer)){
		    	setAnnounce(tracker,1);
			pointer += 7;
			if(isdigit(*pointer)){
				int numchars = strchr(pointer,':')-pointer;
				char*  tempBuff = (char*)malloc(numchars+1);
				tempBuff[numchars]=0;
				strncpy(tempBuff, pointer, numchars);
				pointer += numchars +1;
				int numpeers = atoi(tempBuff)/6;
				free(tempBuff);
				int i = 0;
				for(i = 0; i< numpeers; i++)
				ret += addPeer(((struct Peer*)(pointer+6*i)), &(tracker->peers), &(tracker->peersNum));
				pointer+=numpeers;
			}else if(*pointer == 'l'){
				pointer+=1; //jump l
				struct Peer p;
				while(*pointer !='e'){
					getPeer(&p, (char**)&pointer);
					ret += addPeer(&p, &(tracker->peers), &(tracker->peersNum));
				}
				pointer++; //jump e
			}
	    }else if(startsWith(pointer, "8:complete", end-pointer)){
		pointer += 10;
		pointer += decodeInt(&seed,pointer);
	    }else if(startsWith(pointer, "10:incomplete",end-pointer)){
		pointer += 13;
		pointer += decodeInt(&leech,pointer);
	    }else if(startsWith(pointer, "10:tracker id", end-pointer)){
		    pointer += 13;
		    int numchars = strchr(pointer,':')-pointer;
		    char*  tempBuff = (char*)malloc(numchars+1);
		    tempBuff[numchars]=0;
		    strncpy(tempBuff, pointer, numchars);
		    pointer += numchars +1;
		    int len = atoi(tempBuff);
		    if(!tracker->trackerid){
			    tracker->trackerid = (char*)malloc(len+1);
			    tracker->trackerid[len]=0;
		    }else if(strncmp(tracker->trackerid,pointer, len)){
			    tracker->trackerid = (char*)realloc(tracker->trackerid,len+1);
			    tracker->trackerid[len]=0;
		    }else
			    continue;
		    strncpy(tracker->trackerid, pointer, len);
		    pointer+=len;
		    free(tempBuff);
	    }else
		    pointer++;
    }
	if(seed == -1 ){
		if(option & OPT_SCRAPE){
			printf("No complete field in scrape response\n");
			setScrapeSL(tracker,0);
		}else
			setAnnounceSL(tracker, 0);
	}else{
		tracker->seeders = seed;
		if(option & OPT_SCRAPE)
			setScrapeSL(tracker,1);
		else
			setAnnounceSL(tracker,1);
	}
	
	if(leech == -1){
		if(option & OPT_SCRAPE)
				printf("No incomplete field in scrape response\n");
	}else
		tracker->leechers = leech;
    return ret;
}




double timestamp() {
    struct timeval tv;
    struct timezone tz;
    if(gettimeofday(&tv,&tz) <0)
        fprintf(stderr, "ERROR: timestamp error!\n");
    return (double)tv.tv_sec + (double)tv.tv_usec/1000000.0;
}


void drawBar(int part, int tot, const char* pre) {
    printf("%s [", pre);
    double perc =(double)part/(double)tot;
    if(tot) {
        int filledSpaces = BARLENGTH*perc;
        int i;
        for(i = 0; i < filledSpaces-1; i++)
            putchar('=');
        if(i)
            putchar('>');
        for(i=0; i < BARLENGTH - filledSpaces; i++)
            putchar(' ');
    }
    printf("] %.2f\% (%d/%d)", perc*100.0, part, tot);
    fflush(stdout);
}



inline void printUsage() {
    printf("PeersHarvester v0.1\n\n");
    printf("Usage: peersharvester -[dhn] [-p percent] [-t threshold] [-c count] [-m logtype] -l logfilename torrentfilename\n\n");
    printf("-h\tShow this help message\n");
    printf("-d\tDisable descriptions in logfile\n");
    printf("-n\tNon compact mode\n");
    printf("-p percent\tPeersharvester continues to query trackers until the peers data retrieved reach the percent of total swarm.[Default 50]\n");
    printf("-t threshold\tPeersharvester continues to query trackers until the number of new peers obtained with \"count\" consecutive query is below threshold value.[Default disabled]\n");
    printf("-c count\tSet the number of the consecutive query for the -t option. When -t option is enabled, this value is 10 as default\n");
    printf("-l logfilename\tLog data obtained in logfilename in CSV Format.\n");
    printf("-m logtype\tlogtype can be a string of values in this set: full, tracker,request,ip. This field is set by default to full\n");
    printf("\t\t\tfull = this is the same of \"tracker,request,ip\"\n");
    printf("\t\t\ttracker = log in logfilename informations about trackers founded in the torrent file\n");
    printf("\t\t\trequest = log in logfilename informations about the requests of announce for tracker\n");
    printf("\t\t\tip = log in logfilename the addresses and port number of the peers retrieved\n");
    printf("torrentfilename\tTorrent file to analyze.\n\n");
}

void logCSVData(){
	int i,j,k;
	if(logmode & LOG_MODE_TRACKER){
		if(description)
			fprintf(logFD, "\"TRACKER\",\"ADDRESS\",\"PROTOCOL\",\"PORT\",\"ANNOUNCE\",\"SEEDERS #\",\"LEECHERS #\",\"SWARM #\", \"SCRAPE\",\"HTTPCODE\",\"STATE\",\"ANNOUNCE OK\",\"SCRAPE OK\",\"ANNOUNCE SL\",\"SCRAPE SL\",\"FAILURE\"\n");
		for(i = 0; i < trackersNum; i++)
			csvPrintTracker(trackers+i, logFD);
		fprintf(logFD,"\n");
		fflush(logFD);
	}
	if(logmode & LOG_MODE_REQUEST) {
		if(description)
			fprintf(logFD, "\"TRACKER\",\"REQUEST #\",\"PEERS #\",\"TOTAL #\",\"SECONDS\"\n");
		for(i = 0; i < requestsNum; i++)
			fprintf(logFD, "\"%s:%d\",%d,%d,%d,\"%.4f\"\n", requests[i].tracker->hostname, requests[i].tracker->port, requests[i].num, requests[i].newpeers, requests[i].total, requests[i].time);
		fflush(logFD);
		fprintf(logFD, "\n");
	}
	if(logmode & LOG_MODE_IPADDRS) {
		struct Peer * pointe;
		struct in_addr ip;
		if(description)
			fprintf(logFD,"\n\n\"PEER IP ADDRESS\",\"PORT #\"%s%s\n", (compact?"":",\"PEER ID\""), (logmode & LOG_MODE_TRACKER)?",\"TRACKERS\"":"");
		struct PeerInfo *peersbuf=0;
		int peersNumBuf=0;
		for(i = 0; i < trackersNum; i++)
			for(j = 0; j < trackers[i].peersNum; j++)
				addPeerToPeerInfo(&peersbuf, trackers[i].peers + j, trackers+i);
		char peerid[61];
		printCSVStructure(logFD, peersbuf, compact, logmode & LOG_MODE_TRACKER);
		fflush(logFD);
	}
}


void Cleanup(int signo){
	logCSVData();
	exit(1);
}


int main(int numArgs, char* args[]) {
    int i,opt, pc=50, threshold=0, thresholdCount=10;
    FILE *f, *extlogfile;
    // Check arguments
    if(numArgs < 2) {
        fprintf(stderr, "ERROR: Too few arguments\n");
        printUsage();
        exit(1);
    }
    //Default initialization of log mode to full (tracker, requests and ip)
    logmode = 7;
    char* logFileName = "/dev/null";
    while((opt = getopt(numArgs, args,"hdnp:t:l:m:c:"))!=-1) {
        switch(opt) {
        case 'h':
            printUsage();
            exit(0);
	    break;
	case 'd':
		description=0;
		break;
        case 'p':

            pc = atoi(optarg);
            if(pc <=0 && pc  >100) {
                fprintf(stderr, "ERROR: -p value error %s. -p option value need to be an integer between 0 and 100\n", optarg);
                printUsage();
                exit(1);
            }
            break;
        case 't':
            threshold = atoi(optarg);
            if(threshold <= 0) {
                fprintf(stderr, "ERROR: -t value error %s. -t option value need to be a positive integer\n", optarg);
                printUsage();
                exit(1);
            }
            break;
        case 'l':
		logFileName = (char*) calloc(1,strlen(optarg +1));
		memcpy(logFileName, optarg, strlen(optarg));
            break;
	case 'n':
		compact=0;
		break;
        case 'm':
            logmode = 0;
            char * pointer = optarg;
            char * comma;
            while(*pointer != 0) {
                comma = strchr(pointer,',');
                if(comma)
                    *comma = 0;
                if(!strcmp(pointer,"full"))
                    logmode |= 7;
                else if (!strcmp(pointer,"tracker"))
                    logmode |= LOG_MODE_TRACKER;
                else if (!strcmp(pointer, "request"))
                    logmode |= LOG_MODE_REQUEST;
                else if (!strcmp(pointer, "ip"))
                    logmode |= LOG_MODE_IPADDRS;
                /*else if (!strcmp(pointer, "verbose"))
		logmode |= LOG_MODE_REQUEST_VERBOSE;*/
                else
                    fprintf(stderr,"WARNING: unknown value %s. -m option accept only values in set: full, tracker, request, ip. Remember that no space are allowed\n", pointer);
                pointer+=strlen(pointer);
                if(comma)
                    pointer++;
            }
            break;
	    case 'c':
		    if(!threshold){
			    fprintf(stderr, "WARNING: -t option must be enabled in order to use -c option. Make sure of the -t option is before -c option\n");
			    break;
		    }
		    thresholdCount=atoi(optarg);
	            break;
        }
    }
    double  percent = ((double)pc)/100.0;
    signal(SIGTERM,Cleanup);
    signal(SIGINT,Cleanup);
    logFD = fopen(logFileName,"w");
    extlogfile = fopen("peersharvester.log","a");
    if(!logFD) {
	    fprintf(stderr, "ERROR: while opening file %s for log\n", optarg);
	    printUsage();
	    exit(1);
    }
    /*
    * Open a file descriptor and copy the file content 
    * in a buffer
    */
    if(!(f = fopen(args[optind], "r"))) {
        fprintf(stderr, "ERROR: Opening file %s\n", args[optind]);
        exit(1);
    }
    fseek(f,0,2);
    // Get the torrent file length
    metaInfoFileLength = ftell(f);
    rewind(f);
    // Allocate memory for torrent file content...
    if(!(metaInfoFile	= (char*)malloc(metaInfoFileLength))) {
        fprintf(stderr, "ERROR: Cannot store whole file in memory\n");
        exit(1);
    }
    //... and store it.
    if(fread(metaInfoFile,1,metaInfoFileLength,f)!=metaInfoFileLength)
        fprintf(stderr,"WARNING: File read error: too few bytes red\n");
    fclose(f);
    /*
    * Parse the buffer in order to find tracker urls
    */
    getTrackerURL();
     /*
    * Compute urlencoded hash of sha1 infohash 
     */
    char* infoHash = (char*)malloc(20);
    if(!infoHash) {
	    fprintf(stderr, "ERROR: No memory for info hash\n");
	    return 0;
    }
    getInfoHash(infoHash);
    hashEncoded = (char*)malloc(3*20 +1);
    if(!hashEncoded) {
	    fprintf(stderr, "ERROR: No memory for hash encoded\n");
	    free(infoHash);
	    return 0;
    }
    memset(hashEncoded,0,3*20+1);
    urlEncode(hashEncoded,infoHash, 20);
    free(infoHash);
    //Free metainfo file buffer memory
    free(metaInfoFile);

    /*
    * Check the result of the search
    */
    if(trackersNum <= 0) {
        fprintf(stderr, "ERROR: No announce found in file %s.\n", args[numArgs-1]);
        exit(1);
    }
    /*
    * Print announce tracker list
    */
    printf("Tracker list:\n");
    for(i = 0; i < trackersNum; i++) {
	    switch(getProtocol((trackers+i))) {
        case TRACKER_PROTO_UNKNOWN:
            printf("unknown://");
            break;
        case TRACKER_PROTO_NONE:
            printf("none://");
            break;
        case TRACKER_PROTO_HTTP:
            printf("http://");
            break;
        case TRACKER_PROTO_UDP:
            printf("udp://");
	    break;
	case TRACKER_PROTO_DHT:
		printf("dht://");
        }
        printf("%s:%d\n",(trackers+i)->hostname, (trackers+i)->port);
    }

    /*
    * Connect to trackers
    */
    struct sockaddr_in remote;  // Tracker socket address
    char *getRequest, *trackerResponse;
    if(!(getRequest = (char*)malloc(500))) {
        fprintf(stderr, "ERROR: No memory for Get Request\n");
        exit(1);
    }
    trackerResponse = calloc(1,RESPONSE_SIZE);
    if(!trackerResponse) {
        fprintf(stderr, "ERROR: No memory for HTTP Tracker Response\n");
        exit(1);
    }

    // Socket Descriptor And max syn count in handshake phase
    int sd, synCnt = MAXSYNCNT;



    /*
    * Scrape Phase
    */
    for(i = 0; i < trackersNum; i++) {
	    if(checkForScrape(trackers+i)){
		if(getProtocol(trackers +i) == TRACKER_PROTO_HTTP &&  initTrackerAddress(&remote, trackers+i)) {
			sd = socket(AF_INET, SOCK_STREAM,IPPROTO_TCP);
			setsockopt(sd, IPPROTO_TCP,TCP_SYNCNT, &synCnt, sizeof(synCnt));
			int a =0;
			int b = 2;
			setsockopt(sd, SOL_SOCKET, SO_RCVTIMEO, &b, sizeof(b));
			setsockopt(sd, SOL_SOCKET,SO_KEEPALIVE, &a, sizeof(a));
			setsockopt(sd, SOL_TCP,TCP_KEEPINTVL, &synCnt, sizeof(synCnt));
			setsockopt(sd, SOL_TCP,TCP_KEEPCNT, &synCnt, sizeof(synCnt));
			
			printf("\nScraping... %s:%d\n", (trackers+i)->hostname, (trackers+i)->port);
			if(!connect(sd,(struct sockaddr*)&remote, sizeof(remote))) {
				setState((trackers+i),1);
				int scrape;
				if(scrape = checkForScrape(trackers+i)) {
					makeGETRequest(getRequest,trackers +i,OPT_SCRAPE);
					if(send(sd, getRequest, strlen(getRequest),0) > 0) {
						int bytesReceived;
						memset(trackerResponse,0,RESPONSE_SIZE);
						if((bytesReceived = recv(sd, trackerResponse,RESPONSE_SIZE,MSG_WAITALL))>0){
							parseResponse(trackerResponse, bytesReceived, trackers+i, OPT_SCRAPE);
						}else{
							fprintf(stderr, "WARNING: No HTTP response from tracker %s:%d\n", (trackers+i)->hostname, (trackers+i)->port);
						}
					}
				printTracker(trackers+i, stdout);
				}
			} else {
				setState((trackers+i), 0);
				fprintf(stderr, "WARNING: Connecting error for the tracker: %s:%d\n", (trackers+i)->hostname, (trackers+i)->port);
			}

			close(sd);
		} else if(getProtocol(trackers +i) != TRACKER_PROTO_HTTP) {
			switch(getProtocol(trackers+i)) {
			case TRACKER_PROTO_UDP:
				fprintf(stderr,"WARNING: UDP protocol not yet implemented. Tracker %s:%d skipped\n", (trackers+i)->hostname, (trackers+i)->port);
				setState((trackers+i),0);
				break;
			case TRACKER_PROTO_DHT:
				fprintf(stderr,"WARNING: DHT protocol not yet implemented. Tracker %s:%d skipped\n", (trackers+i)->hostname, (trackers+i)->port);
				setState((trackers+i),0);
				break;
			case TRACKER_PROTO_UNKNOWN:
				fprintf(stderr,"WARNING: Unknown protocol. Tracker %s:%d skipped\n", (trackers+i)->hostname, (trackers+i)->port);
				setState((trackers+i),0);
				break;
			case TRACKER_PROTO_NONE:
				fprintf(stderr,"WARNING: Protocol not found. Tracker %s:%d skipped\n", (trackers+i)->hostname, (trackers+i)->port);
				setState((trackers+i),0);
				break;
			}
		}
	    }
	}
    if (logmode & LOG_MODE_REQUEST || logmode & LOG_MODE_IPADDRS){
	/*
	* Peer Harvesting Phase
	*/
	double t1,t2;
	char * tempRequest = (char*)malloc(15);
	int allocated = REQNUM;
	if(logmode & LOG_MODE_REQUEST)
		requests = (struct Request *)malloc(sizeof(struct Request)*(allocated));
	
	for(i = 0; i < trackersNum; i++) {
		if(isAlive(trackers +i) && initTrackerAddress(&remote, trackers+i)) {
			printf("Harvesting... %s:%d\n", (trackers+i)->hostname, (trackers+i)->port);
			int reqCount = 0, tc=0, peersCount=0, swarm=0;
			makeGETRequest(getRequest,trackers +i,(compact?OPT_ANNOUNCE_START|OPT_COMPACT:OPT_ANNOUNCE_START));
			do{
				reqCount++;
				sd = socket(AF_INET, SOCK_STREAM,IPPROTO_TCP);
				setsockopt(sd, IPPROTO_TCP,TCP_SYNCNT, &synCnt, sizeof(synCnt));
				t1 = timestamp();
				if(!connect(sd,(struct sockaddr*)&remote, sizeof(remote))) {
					if(send(sd, getRequest, strlen(getRequest),0) > 0) {
						int bytesReceived;
						memset(trackerResponse,0,RESPONSE_SIZE);
						if((bytesReceived = recv(sd, trackerResponse,RESPONSE_SIZE,MSG_WAITALL))>0) {
							close(sd);
							t2 = timestamp();
							int result = parseResponse(trackerResponse, bytesReceived, trackers+i, (compact?OPT_COMPACT:0));
							if(result >= 0)
								peersCount = result;
							else if (result <0 && !compact){
								reqCount = 0;
								compact = 1;						
								printf("WARNING: Compact mode enabled\n");
								swarm = ((trackers+i)->leechers + (trackers+i)->seeders);
								makeGETRequest(getRequest,trackers +i,OPT_ANNOUNCE_START|OPT_COMPACT);
								continue;
							}else
								break;
							if(logmode & LOG_MODE_REQUEST){
								// New Request
								requestsNum++;
								if(requestsNum > allocated){
									allocated += REQNUM;
									requests = (struct Request *)realloc(requests, sizeof(struct Request)*(allocated));
								}
								(requests+requestsNum-1)->tracker = (trackers + i);
								(requests+requestsNum-1)->num = reqCount;
								(requests+requestsNum-1)->newpeers = peersCount;
								(requests+requestsNum-1)->total = trackers[i].peersNum;
								(requests+requestsNum-1)->time = t2 - t1;
							}
							if(threshold){
								if(peersCount < threshold) tc++;
								else tc = 0;
							}
							if(threshold &&  tc >= thresholdCount) {
								printf("\nInterrupted: %d consecutive values less than %d\n", tc,threshold);
								int kx = 0;
								time_t rawtime;
								time ( &rawtime );
								fprintf (extlogfile, "%s,", ctime (&rawtime) );
								for(; kx < numArgs; kx ++)
									fprintf(extlogfile, "%s ", args[kx]);
								fprintf(extlogfile, ", Interrupted due to threshold after %d requests\n", reqCount);
								break;
							} else {
								sprintf(tempRequest,"\rRequest #%d", reqCount);
								drawBar(trackers[i].peersNum,((trackers+i)->leechers + (trackers+i)->seeders), tempRequest);
							}
							
						}else{
							fprintf(stderr, "\nWARNING: No HTTP response from tracker %s:%d\n", (trackers+i)->hostname, (trackers+i)->port);
							int kx = 0;
							time_t rawtime;
							time ( &rawtime );
							fprintf (extlogfile, "%s,", ctime (&rawtime) );
							for(; kx < numArgs; kx ++)
								fprintf(extlogfile, "%s ", args[kx]);
							fprintf(extlogfile, ", No HTTP response after %d requests\n", reqCount);
							break;
						}
					}
				} else {
					int kx = 0;
					time_t rawtime;
					time ( &rawtime );
					fprintf (extlogfile, "%s,", ctime (&rawtime) );
					for(; kx < numArgs; kx ++)
						fprintf(extlogfile, "%s ", args[kx]);
					fprintf(extlogfile, ", Connecting Error after %d requests\n", reqCount);
					fprintf(stderr, "WARNING: Connecting error for the tracker %s\n", (trackers+i)->hostname);
					close(sd);
					break;
				}
				swarm = ((trackers+i)->leechers + (trackers+i)->seeders);
			}while(swarm &&   (trackers[i].peersNum<swarm*percent));
			if((swarm && trackers[i].peersNum>=swarm*percent) || (threshold &&  tc >= thresholdCount)){
				sd = socket(AF_INET, SOCK_STREAM,IPPROTO_TCP);
				setsockopt(sd, IPPROTO_TCP,TCP_SYNCNT, &synCnt, sizeof(synCnt));
				makeGETRequest(getRequest,trackers +i,OPT_ANNOUNCE_STOP); 
				if(!connect(sd,(struct sockaddr*)&remote, sizeof(remote))) 
					send(sd, getRequest, strlen(getRequest),0) ;
				close(sd);
				
			}
			if(!swarm){
				fprintf(stderr, "\nWARNING: No peers found on tracker: %s:%d\n", (trackers+i)->hostname, (trackers+i)->port);
				int kx = 0;
				time_t rawtime;
				time ( &rawtime );
				fprintf (extlogfile, "%s,", ctime (&rawtime) );
				for(; kx < numArgs; kx ++)
					fprintf(extlogfile, "%s ", args[kx]);
				fprintf(extlogfile, ", No Peers found\n");
			}
			if(!(trackers[i].peersNum<swarm*percent)){
				int kx = 0;
				time_t rawtime;
				time ( &rawtime );
				fprintf (extlogfile, "%s,", ctime (&rawtime) );
				for(; kx < numArgs; kx ++)
					fprintf(extlogfile, "%s ", args[kx]);
				fprintf(extlogfile, ", Reached the percentage\n");				
			}
		}
		printf("\n");
	}
	free(tempRequest);
	free(hashEncoded);
	free(getRequest);
	free(trackerResponse);
    }

    /*
    *	Log phase
    */
    logCSVData();
    fclose(logFD);
    fclose(extlogfile);
    free(trackers);
    printf("\n\n");
    return 0;
}

