/***************************************************************************
 *   Copyright (C) 2006 by BardaK   *
 *   bardak@BardaX   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "trackerUtils.h"
#include <unistd.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <time.h>



struct Tracker* newTracker(struct Tracker * rep, unsigned int * size){
	struct Tracker * ref  = rep;
	if(*size >0){
		*size+=1;
		rep = realloc(rep,sizeof(struct Tracker)*(*size));
		ref =rep + *size-1;		
	}else if(*size == 0){
		*size+=1;
		rep = (struct Tracker*)malloc(sizeof(struct Tracker));
		if(!rep){
			fprintf(stderr, "ERROR: no memory for trackers list\n");
			exit(1);
		}
		ref = rep;
	}
	else {
		fprintf(stderr, "BUG: size passed newTracker function less than 0\n");
		return 0;
	}
	ref->hostname = 0;
	ref->port = 0;
	ref->announce=0;
	ref->seeders = 0;
	ref->leechers = 0;
	ref->httpCode = 0;
	ref->trackerid = 0;
	ref->peersNum=0;
	ref->flags=0;
	ref->failure=0;
	return rep;	
}

int initTracker(struct Tracker * tracker, const char* url, unsigned int length){
	if(!strncmp(url, "http://",7)){
		setProtocol(tracker,TRACKER_PROTO_HTTP);
		tracker->port = 80;
		 url+=7;
		 length -=7;
	}
	else if(!strncmp(url, "udp://",6)){
		setProtocol(tracker, TRACKER_PROTO_UDP);
		 url+=6;
		 length -=6;
	}else if(!strncmp(url, "dht://",6)){
		setProtocol(tracker, TRACKER_PROTO_DHT);		
		url+=6;
		length -=6;
	}else{
		fprintf(stderr, "BUG: unknown protocol for tracker ");
		fwrite( url, length,1,stderr);
		fprintf(stderr, "\n");
		setProtocol(tracker, TRACKER_PROTO_UNKNOWN);
		return 0;
	}
	
	const char* portIndex = url;
	const char* pathIndex = url;
	while(portIndex < url+length && *portIndex != ':')
		portIndex++;
	if(portIndex == url + length && *portIndex != ':') portIndex = 0;
	while(pathIndex < url+length && *pathIndex != '/')
		pathIndex++;
	if(pathIndex == url + length && *pathIndex != '/') pathIndex = 0;
	int len = (portIndex)?portIndex-url:0;
	if(!len)
		len = (pathIndex)?pathIndex-url:0;
	if(!len)
		len = length;
	tracker->hostname = (char*)malloc(len+1);
	if(!tracker->hostname){
		fprintf(stderr, "ERROR: No memory for tracker url\n");
		exit(1);
	}
	tracker->hostname[len] = 0;
	strncpy(tracker->hostname, url, len);	
	if(portIndex){		
		portIndex++;
		tracker->port = 0;
		while(isdigit(*portIndex)){
			tracker->port *=10;
			tracker->port += *portIndex -'0';
			portIndex++;
		}
	}
	if(pathIndex){
		len = url+length-pathIndex;
		tracker->announce = (char*)malloc(len+1);
		if(!tracker->announce){
			fprintf(stderr, "ERROR: No memory for announce\n");
			exit(1);
		}
		tracker->announce[len]=0;
		strncpy(tracker->announce, pathIndex, len);
	}else{
		tracker->announce = (char *) malloc(2);		
		if(!tracker->announce){
			fprintf(stderr, "ERROR: No memory for announce\n");
			exit(1);
		}
		tracker->announce[1]=0;
		tracker->announce[0]='/';
	}
	setState(tracker,1);
	return 1;
}

int checkForScrape(struct Tracker * tracker) {
	char * pointer = tracker->announce;
	if(!pointer) return 0;
	int i;
	for(i = 0; i < strlen(tracker->announce); i++)
		if(*(tracker->announce+i) == '/') pointer = tracker->announce +i;
	if(!strncmp((pointer+1), "announce",8))
		return 1;
	return 0;
}

int getScrapePath(struct Tracker * tracker, char* scrape){
	// Check if scape is supported by the URL;
	if(!checkForScrape(tracker))
		return 0;
	char * pointer = scrape;
	int i;
	// Copy first part of announce's path. Example: /tracker/announce?a=10&b=5
	//                                                   from here->|            |<- to here
	// Result is:  scrape => "/tracker/"
	for (i = 0; strncmp((tracker->announce+i),"announce",8); i++)
		scrape[i] = tracker->announce[i];
	// Append "scrape" to the string. Result is: scrape => "/tracker/scrape"
	memcpy((scrape+i), "scrape",6);
	i+=6;
	
	// Copy the rest of announce's path. Result is: scrape => "/tracker/scrape?a=10&b=5"
	for(;tracker->announce[ i + 2 ] != 0; i++)
		scrape[i] = tracker->announce[ i + 2 ];
		
	scrape[i] = 0;
	return 1;
}

void printTracker(struct Tracker * tracker, FILE* fd){
	char * proto =  (char*)malloc(10);
	char*id=(char*)malloc(50);
	switch (getProtocol(tracker)){
		case TRACKER_PROTO_NONE:
			strcpy(proto, "None");
			break;
		case TRACKER_PROTO_HTTP:
			strcpy(proto, "HTTP");
			break;
		case TRACKER_PROTO_UDP:
			strcpy(proto, "UDP");
			break;
		case TRACKER_PROTO_DHT:
			strcpy(proto, "DHT");
			break;
		case TRACKER_PROTO_UNKNOWN:
			strcpy(proto, "Unknown");
			break;
	}
	strcpy(id,((tracker->trackerid)?(tracker->trackerid):("Unknown")));
	fprintf(fd,"Tracker id(%s), address (%s), protocol (%s), port (%d), announce (%s), seeders (%d), leechers(%d), swarm(%d), scrape (%s), httpCode(%d), state (%s)\n", id, tracker->hostname, proto, tracker->port, tracker->announce, tracker->seeders, tracker->leechers, tracker->seeders+tracker->leechers, checkForScrape(tracker)?"Supported":"Not supported", tracker->httpCode, (isAlive(tracker))?"UP":"DOWN");
	free(proto);
	free(id);
}

void printTrackerAddress(FILE* fd, struct Tracker* tracker){
	char * proto =  (char*)malloc(10);
	switch (getProtocol(tracker)){
		case TRACKER_PROTO_NONE:
			strcpy(proto, "None");
			break;
		case TRACKER_PROTO_HTTP:
			strcpy(proto, "http");
			break;
		case TRACKER_PROTO_UDP:
			strcpy(proto, "udp");
			break;
		case TRACKER_PROTO_DHT:
			strcpy(proto, "dht");
			break;
		case TRACKER_PROTO_UNKNOWN:
			strcpy(proto, "unknown");
			break;
	}
	fprintf(fd, "%s://%s:%d", proto, tracker->hostname, tracker->port);
}

void csvPrintTracker(struct Tracker* tracker, FILE*fd){
	char * proto =  (char*)malloc(10);
	char*id=(char*)malloc(50);
	switch (getProtocol(tracker)){
		case TRACKER_PROTO_NONE:
			strcpy(proto, "None");
			break;
		case TRACKER_PROTO_HTTP:
			strcpy(proto, "HTTP");
			break;
		case TRACKER_PROTO_UDP:
			strcpy(proto, "UDP");
			break;
		case TRACKER_PROTO_DHT:
			strcpy(proto, "DHT");
			break;
		case TRACKER_PROTO_UNKNOWN:
			strcpy(proto, "Unknown");
			break;
	}
	strcpy(id,((tracker->trackerid)?(tracker->trackerid):("Unknown")));
	fprintf(fd,"\"%s\", \"%s\", \"%s\", %d, \"%s\", %d, %d, %d, \"%s\", %d, \"%s\", \"%s\", \"%s\", \"%s\", \"%s\",\"%s\"\n", 
		  id, 
		  tracker->hostname, 
		  proto, 
		  tracker->port, 
		  tracker->announce, 
		  tracker->seeders, 
		  tracker->leechers, 
		  tracker->seeders+tracker->leechers,
		  checkForScrape(tracker)?"Supported":"Not supported",
		  tracker->httpCode,
		  isAlive(tracker)?"UP":"DOWN",
		  tracker->flags&0x2?"YES":"NO",
		  tracker->flags&0x4?"YES":"NO",
		  tracker->flags&0x8?"YES":"NO",
		  tracker->flags&0x10?"YES":"NO",
	          tracker->failure);
}
/*
TRACKER FLAGS:
bit		Description
0		state
1		announce ok
2		scrape ok
3		announce seeders/leechers
4		scrape	seeders/leechers
5-7		protocol
*/
void setState(struct Tracker* tracker, unsigned char value){
	tracker->flags &=0xFE;
	tracker->flags |= (value&0x1);
}

inline unsigned char isAlive(struct Tracker* tracker){
		return tracker->flags&0x01;
}

void setAnnounce(struct Tracker* tracker, unsigned char value){	
	tracker->flags &=0xFD;
	tracker->flags |= (value&0x1)<<1;
}

void setScrape(struct Tracker* tracker, unsigned char value){
	tracker->flags &=0xFB;
	tracker->flags |= (value&0x1)<<2;
}

void setAnnounceSL(struct Tracker* tracker, unsigned char value){
	tracker->flags &=0xF7;
	tracker->flags |= (value&0x1)<<3;
}

void setScrapeSL(struct Tracker* tracker, unsigned char value){
	tracker->flags &=0xEF;
	tracker->flags |= (value&0x1)<<4;
}

void setProtocol(struct Tracker* tracker, unsigned char value){	
	tracker->flags &=0x1F;
	tracker->flags |= (value&7)<<5;;
}

inline char getProtocol(struct Tracker* tracker){
	return (tracker->flags&0xE0)>>5;
}


void freeTracker(struct Tracker* tracker){
	if(tracker->hostname)
		free(tracker->hostname);
	if(tracker->announce)
		free(tracker->announce);
	if(tracker->trackerid)
		free(tracker->trackerid);
	int i;
	for(i=tracker->peersNum-1; i>=0; i++)
		free(tracker->peers +i);
	if(tracker->failure)
		free(tracker->failure);
	free(tracker);
}

int comparePeers(struct Peer* c, struct Peer*b){
	int i = 0;
	if(c->ip > b->ip)
		return 1;
	else if (c->ip < b->ip)
		return -1;
	if(c->port > b->port)
		return 1;
	else if(c->port < b->port)
		return -1;
	return 0;
}

int addPeer(struct Peer * newpeer, struct Peer* *peers, unsigned int *peersNum) {
	int i = 0;
	struct Peer* p = *peers;	
	for(i=0; i <*peersNum; i++){
		if(!memcmp(newpeer, (char*)p+i*sizeof(struct Peer),6))
			break;
	}
	if (i==*peersNum){
		if(*peersNum == 0)
				*peers = (struct Peer*)malloc(sizeof(struct Peer)*MAX_PEERS_ALLOCATION);
		else if(*peersNum % MAX_PEERS_ALLOCATION == 0)
			*peers = realloc(*peers, sizeof(struct Peer)*(*peersNum+MAX_PEERS_ALLOCATION));
		memcpy((*peers+*peersNum),newpeer,sizeof(struct Peer));
		*peersNum+=1;
		return 1;
	}
	return 0;
	/*	
	srand(time(0));
	int r = (int)(rand()*(*peersNum)/RAND_MAX);
		for(i=r; i <*peersNum; i++){
			if(!memcmp(newpeer, (char*)p+i*sizeof(struct Peer),6))
				break;
		}
		if(i == *peersNum) {
			for(i=0; i< r; i++)
				if(!memcmp(newpeer, (char*)p+i*sizeof(struct Peer),6))
					break;
		}
		if(i==r){
			if(*peersNum == 0)
				*peers = (struct Peer*)malloc(sizeof(struct Peer)*MAX_PEERS_ALLOCATION);
			else if(*peersNum % MAX_PEERS_ALLOCATION == 0)
				*peers = realloc(*peers, sizeof(struct Peer)*(*peersNum+MAX_PEERS_ALLOCATION));
			memcpy((*peers+*peersNum),newpeer,sizeof(struct Peer));
			*peersNum+=1;
			return 1;
		}
		return 0;
	*/
}



struct Peer* getPeer(struct Peer* dest,  char** buffer){
	int i;
	char *pointer = *buffer;
	char *end = pointer + bencodeSkipChars(pointer);
	memset(dest,0,sizeof(struct Peer));
	char* tempBuff = (char*)malloc(255);
	*pointer++; //jump d
	while(*pointer!='e'){
		if(startsWith(pointer, "7:peer id",end-pointer)){
			pointer+=9;
			dest->peerid[20] = 0;
			pointer += decodeString(dest->peerid, pointer);
		}else if(startsWith(pointer, "2:ip", end-pointer)){
			pointer+=4;
			memset(tempBuff,0,255);
			pointer += decodeString(tempBuff, pointer);
			struct in_addr addr;
			inet_aton(tempBuff, &addr);
			dest->ip = addr.s_addr;
		}else if (startsWith(pointer, "4:port",end-pointer)){
			pointer += 6;
			int port;
			pointer += decodeInt(&port, pointer);
			dest->port = htons(port);
		}else
			pointer++;
	}
	pointer ++; //jump e
	*buffer = pointer;
	return dest;
}





