/***************************************************************************
 *   Copyright (C) 2006 by BardaK   *
 *   bardak@BardaX   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef _TRACKER_UTILS_H_
#define _TRACKER_UTILS_H_

#include <string.h>
#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>

#define TRACKER_PROTO_UNKNOWN	7
#define TRACKER_PROTO_NONE	0
#define TRACKER_PROTO_HTTP	1
#define TRACKER_PROTO_UDP		2
#define TRACKER_PROTO_DHT		3


#define MAX_PEERS_ALLOCATION 50



struct Peer {
    int ip;
    short port;
    char peerid[21];
};


struct Tracker {
	char* hostname;
	unsigned short port;
	char* announce;
	unsigned int seeders;
	unsigned int leechers;
	unsigned int httpCode;
	char* trackerid;
	struct Peer * peers;
	unsigned int peersNum;
	unsigned char flags;
	char* failure;
};

/*
TRACKER FLAGS:
bit		Description
0		state
1		announce ok
2		scrape ok
3		announce seeders/leechers
4		scrape	seeders/leechers
5-7		protocol
*/

struct Tracker* newTracker(struct Tracker * rep, unsigned int* size);
int initTracker(struct Tracker * tracker, const char* url, unsigned int length);
int checkForScrape(struct Tracker * tracker) ;
int getScrapePath(struct Tracker * tracker, char* scrape);
void printTracker(struct Tracker * tracker, FILE* fd);
void csvPrintTracker(struct Tracker* tracker, FILE*fd);
void setState(struct Tracker* tracker, unsigned char value);
inline unsigned char isAlive(struct Tracker* tracker);
void setAnnounce(struct Tracker* tracker, unsigned char value);
void setScrape(struct Tracker* tracker, unsigned char value);
void setAnnounceSL(struct Tracker* tracker, unsigned char value);
void setScrapeSL(struct Tracker* tracker, unsigned char value);
void setProtocol(struct Tracker* tracker, unsigned char value);
inline char getProtocol(struct Tracker*);
void freeTracker(struct Tracker* tracker);
int comparePeers(struct Peer*c, struct Peer*b);
struct Peer* getPeer(struct Peer* dest,  char** buffer);
int addPeer(struct Peer * newpeer, struct Peer* *peers, unsigned int *peersNum);
void printTrackerAddress(FILE* fd, struct Tracker* tracker);

#endif



